/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
 #include "common.h" 
 #include "support.h"
 
 #include <gtk/gtk.h>
 
 


artnet_node node ;
artnet_node_entry active_node ;
int nodes_found = 0;

int ticks_total ;
int ticks ;
fade_status_e fade_status ;
uint8_t data_output[SLIDERS] ;
uint8_t old_data[SLIDERS];
uint8_t data_sliders[SLIDERS] ;
uint8_t data_start[SLIDERS] ;


// hold a 1 is the flash button is down, else 0
int flashes_on[FLASHES] ;


// fade time in milli seconds
int fade_time ;
int active_port ;




/*
 * Generic get slider values function
 *
 */
int get_slider_values(GtkWidget *widget, uint8_t *data, int length) {
	int i;
	gchar *gstr ;
	GtkVScale *scale ;

	for(i=0; i< MIN(length,SLIDERS); i++) {
		gstr = g_strdup_printf("vscale%i", i+1) ;
		scale = (GtkVScale*) lookup_widget(widget ,gstr) ;
		*data = gtk_range_get_value((GtkRange*) scale) ;
		data++ ;
	}

	return MIN(length, SLIDERS) ;	
}

/*
 * Generic set slider values function
 *
 */
int set_slider_values(GtkWidget *widget, uint8_t *data, int length) {
	int i;
	gchar *gstr ;
	GtkVScale *scale ;

	for(i=0; i< MIN(length,SLIDERS); i++) {
		gstr = g_strdup_printf("vscale%i", i+1) ;
		scale = (GtkVScale*) lookup_widget(widget ,gstr) ;
		gtk_range_set_value((GtkRange*) scale, *data) ;
		data++ ;
	}
	return 0;
}


/*
 * Blackout
 *
 */
int blackout(GtkWidget *widget) {
	uint8_t data[SLIDERS] ;

	memset(data, 0x00, SLIDERS) ;

	return set_slider_values(widget,data,SLIDERS) ;	
	
}


/*
 * Full on
 *
 */
int all_on(GtkWidget *widget) {
	uint8_t data[SLIDERS] ;

	memset(data, 0xFF, SLIDERS) ;

	return set_slider_values(widget,data,SLIDERS) ;	
	
}

int get_scene_from_template(beat_scene *scene, uint8_t *data, int length) {
	int i, len, tmp;
	
	len = MIN(sizeof(scene->defn) , length) ;
	
	// use two loops here
	
	// first look for fixed or random values
	for(i=0; i <length; i++) {
		if(scene->defn[i].mode == FIXED) {
			data[i] = scene->defn[i].param1 ;			
		} else if (scene->defn[i].mode == RAND) {
			tmp = rand() % (scene->defn[i].param2 - scene->defn[i].param1) ;
			data[i] = scene->defn[i].param1 + tmp ;	
		}
	}
	
	// now look for clone or inverses
	for(i=0; i <length; i++) {
		if(scene->defn[i].mode == CLONE) {
			if(0< scene->defn[i].param1 && scene->defn[i].param1 <= SLIDERS) {
 				data[i] = data[scene->defn[i].param1-1] ;			
			} else {
				printf("VERY BAD!! PARAM1: %d out of RANGE\n", scene->defn[i].param1 ) ;
			}
		} else if (scene->defn[i].mode == INVERSE) {
			if(0< scene->defn[i].param1 && scene->defn[i].param1 <= SLIDERS) {
 				data[i] = 255 - data[scene->defn[i].param1-1] ;			
			} else {
				printf("VERY BAD!! PARAM1 out of RANGE\n") ;
			}
		}
		
	}
	
	for(i=0; i < length; i++ ) {
		printf(" %d ", data[i]) ;
	}
	printf("\n") ;
	
	return 0;
}

/*
 * Call upon a beat
 *
 */
void beat_mode(GtkWidget *widget) {
	static int beats = 0;
	GtkSpinButton *spin = (GtkSpinButton*) lookup_widget(widget, "spinbutton1") ;
	GtkTreeView *table = (GtkTreeView*) lookup_widget(widget, "treeview2") ;
	GtkListStore *listStore = (GtkListStore*) gtk_tree_view_get_model(table) ;
	GtkTreeSelection *selection = gtk_tree_view_get_selection(table );
	GtkTreeIter iter ;
	int count ;
	static int scene = 0;
	beat_scene tmp ;
	uint8_t data[SLIDERS] ;
	
	count = get_beat_scene_count() ;
	
	if(count ==0)
		return ;
	
	if(beats > gtk_spin_button_get_value_as_int(spin) &&
		gtk_spin_button_get_value_as_int(spin)  != 0) {
		printf("choosing new scene\n") ;
		// pick a new scene here
		scene = rand() % count;
		beats = 0;
			
		gtk_tree_model_iter_nth_child((GtkTreeModel*) listStore, &iter, NULL, scene);
		gtk_tree_selection_select_iter(selection, &iter) ;
	}
	
	printf("going with scene %d\n", scene) ;
	
	get_from_beat(scene, &tmp) ;
	get_scene_from_template(&tmp, data, SLIDERS) ;	
	artnet_send_dmx(node, active_port, SLIDERS, data) ;
	
	// select on listStore
	
	
	beats++ ;
}


/*
 * Called when we enter slow mode
 */
void slow_mode(GtkWidget *widget) {
	uint8_t *data ;
	int len, count ;
	static int scene = 0;
	
	GtkCheckButton *check = (GtkCheckButton*) lookup_widget(widget, "random_scene_button") ;
	GtkTreeView *table = (GtkTreeView*) lookup_widget(widget, "treeview1") ;
	GtkListStore *listStore = (GtkListStore*) gtk_tree_view_get_model(table) ;
	GtkTreeSelection *selection = gtk_tree_view_get_selection(table );
	GtkTreeIter iter ;
	
	count = get_slow_scene_count() ;
	
	if (count ==0) 
		return ;
		
	if(gtk_toggle_button_get_active((GtkToggleButton*) check) ) {
	
		// rand between 0 and count-1
		scene = rand() % count;
	}
	
	gtk_tree_model_iter_nth_child((GtkTreeModel*) listStore, &iter, NULL, scene);
	gtk_tree_selection_select_iter(selection, &iter) ;
	
	data = get_from_slow(scene, &len) ;

	// set dmx values
	artnet_send_dmx(node, active_port, len, data) ;
	
	scene++ ;
	
	if (scene == count)
		scene = 0 ;
}

void audio_callback (audio_events event, void *usr_data) {
	GtkWidget *window = (GtkWidget*) usr_data ;
	
	GtkSpinButton *spin = (GtkSpinButton*) lookup_widget(window, "scene_timeout_button") ;
	static int ticks = 0 ;
	
	ticks++ ;
	
	if(event == AUDIO_BEAT) {
		ticks = 0;
		beat_mode(window) ;
	} 
	
	if(event == AUDIO_SLOW ||
		event == AUDIO_TRANS ||
	 	(ticks > audio_window_get_samples_per_second() * gtk_spin_button_get_value_as_int(spin) &&
			gtk_spin_button_get_value_as_int(spin) != 0)	) {
			
		slow_mode(window);
		ticks = 0;
	}
	
}


artnet_node_entry lookup_node_entry_by_id(artnet_node n, int id) {
	artnet_node_list nl = artnet_get_nl(n) ;
	artnet_node_entry e = artnet_nl_first(nl) ;
	int i = 1;
	
	if(id < 1 || id > MAX_NODES)
		return NULL;
	
	
	while(e != NULL && i <= MAX_NODES) {
		if(i == id)
			break;

		e = artnet_nl_next(nl) ;
		i++ ;
	}
	
	if(i == id && e != NULL) 
		return e ;

	return NULL ;
	
	
}


void change_active_node_to(artnet_node n, GtkWidget *widget, int id) {
	artnet_node_entry e = lookup_node_entry_by_id(n, id) ;
	GtkRadioButton *rbutton ;
	
	gchar *gstr ;
	
	int i ;
	
	if(e != NULL) {
		active_node = e ;
		
		// change this node to match our new peer
		// else we won't get any dmx packets
		artnet_set_subnet_addr(n, e->sub) ;

		for(i=0; i < MIN(e->numbports, ARTNET_MAX_PORTS) ; i++) {
			gstr = g_strdup_printf("port_radiobutton%i", i+1) ;
			rbutton = (GtkRadioButton *) lookup_widget(widget, gstr) ;
			gstr = g_strdup_printf("%i", e->swout[i]) ;
			
			gtk_button_set_label((GtkButton *) rbutton, gstr) ;
			gtk_widget_set_sensitive((GtkWidget *) rbutton, TRUE) ;
			
  			// set the universe address 
  			artnet_set_port_addr(node, i, ARTNET_INPUT_PORT, e->swout[i] & 0x0F) ;
		}
		
		for(i= MIN(e->numbports, ARTNET_MAX_PORTS) ; i < ARTNET_MAX_PORTS; i++) {
			gstr = g_strdup_printf("port_radiobutton%i", i+1) ;
			rbutton = (GtkRadioButton *)  lookup_widget(widget, gstr) ;
			gtk_widget_set_sensitive((GtkWidget *) rbutton, FALSE) ;
		}
		
	}
	
}	

int on_reply_recv(artnet_node n, void *pp, void *data) {
	artnet_node_list nl = artnet_get_nl(n) ;
	
	artnet_node_entry e = artnet_nl_first(nl) ;
	GtkComboBox *combobox1 ;
	
	if(nodes_found != artnet_nl_get_length(nl)) {
		nodes_found++ ;
		combobox1 = (GtkComboBox *) lookup_widget((GtkWidget *) data, "combobox1") ;
		
		e = lookup_node_entry_by_id(n, nodes_found) ;	
		gtk_combo_box_append_text(combobox1, e->shortname) ;
		
		if(nodes_found == 1 )
			gtk_combo_box_set_active(combobox1, 0 ) ;
		
		if(active_node == NULL && artnet_nl_first(nl) != NULL)
			change_active_node_to(n, data, 1) ;

		
	}
	
	return 0;
}


void htp_merge(uint8_t *dst, uint8_t *src, int length) {
	int i ;
	
	for(i=0; i<length; i++) {
		
		*dst = MAX(*dst, *src)  ;
		dst++ ;
		src++ ;
	}
	
	
}

// called approx 36 times per second
// this generates the dmx data
int timeout(gpointer data) {
	
	GtkNotebook *book = (GtkNotebook*) lookup_widget((GtkWidget*) data, "notebook1") ;
	uint8_t flash_data[SLIDERS] ;
	
	if(gtk_notebook_get_current_page(book) != 0) {
		return FALSE ;
	}
	
	artnet_node n = node; 
	int i;
	int temp, length ;
	
	if(fade_status == FADE_START) {
		
		fade_status = IN_FADE ;
		ticks = 0 ;
		ticks_total = fade_time / 1000 * TICKS_PER_SECOND ;	

		printf("starting fade, total ticks is %d\n", ticks_total) ;

		
		for(i=0; i< SLIDERS; i++) {
			data_start[i] = data_output[i] ;
		}	
		printf(" data start[0] is 0x%02hhx\n", data_start[0] ) ;		
		
	} 
	
	if( fade_status == FADE_CHANGE) {
	//	printf("\n\nFADE CHANGE!!!!!!\n") ;
		
		ticks_total = (ticks_total -ticks) * fade_time / 1000 / ticks_total * TICKS_PER_SECOND ;

		if(ticks_total < 1) {
			fade_status = NO_FADE ;
		} else {
			fade_status = IN_FADE;
			ticks = 0;
			for(i=0; i< SLIDERS; i++) {
				data_start[i] = data_output[i] ;
			}	
		}	
	}
	
	if(fade_status == IN_FADE) {
		ticks++ ;
		for(i=0; i < SLIDERS; i++) {
			temp = data_start[i] + (ticks* (data_sliders[i]-data_start[i]) / ticks_total) ;
			assert(temp >= 0 && temp <=255) ;
			data_output[i] = temp ;
		}			
		
		if(ticks == ticks_total) 
			fade_status = NO_FADE ;
	} else {
		memcpy(data_output, data_sliders, SLIDERS) ;
	} 

	memset (flash_data, 0x00, SLIDERS) ;
	
	for(i=0; i<FLASHES ; i++) {
		if (flashes_on[i]) {
			printf("flash %i is on!\n", i) ;
			
			htp_merge(flash_data, get_flash(i, &temp), SLIDERS) ;
		}			
		
	}

	htp_merge(data_output, flash_data, SLIDERS) ;
	
	// work out how much data has changed
	length = 0;
	for(i = SLIDERS-1; i >= 0; i--) {
		if(old_data[i] != data_output[i]) {
			length = i+1 ;
			break ;
		}
	}
	
	
//	printf("length %i, output is 0x%02hhx 0x%02hhx 0x%02hhx 0x%02hhx\n", length, data_output[0], old_data[0], data_output[1], old_data[1] );
	memcpy(old_data, data_output, length) ;
	
	if(length) {
		artnet_send_dmx(n, active_port, SLIDERS, data_output) ;
	}
	return TRUE ;
}

/*
 * Very very bad here, ignoring the length
 * need to figure out how to add columns in a for loop
 *
 */
void add_slow_scene_to_table(GtkWidget* widget, uint8_t *data, int length) {
	GtkTreeView *table = (GtkTreeView*) lookup_widget(widget, "treeview1") ;
	GtkListStore *listStore = (GtkListStore*) gtk_tree_view_get_model(table) ;
	GtkTreeIter iter ;
	
	gtk_list_store_append(listStore, &iter) ;
	gtk_list_store_set(listStore, &iter, COLUMN_1, data[0],
										 COLUMN_2,data[1],
										 COLUMN_3,data[2],
										 COLUMN_4,data[3],
										 COLUMN_5,data[4],
										 COLUMN_6,data[5],
										 COLUMN_7,data[6],
										 COLUMN_8,data[7],
										 COLUMN_9,data[8],
										 COLUMN_10,data[9],
										 COLUMN_11,data[10],
										 COLUMN_12,data[11],
										 -1 );
}


void add_beat_scene_to_table(GtkWidget* widget, beat_scene *scene) {
	GtkTreeView *table = (GtkTreeView*) lookup_widget(widget, "treeview2") ;
	GtkListStore *listStore = (GtkListStore*) gtk_tree_view_get_model(table) ;
	GtkTreeIter iter ;
	gchar *gstr = g_strdup_printf("%s",scene->name) ;
	
	printf("in add beat scene\n") ;
	
	gtk_list_store_append(listStore, &iter) ;
	gtk_list_store_set(listStore, &iter, COLUMN_1, gstr, -1 );	
	
}


int show_beat_window(GtkWidget *beat_window, beat_scene scene) {
	gtk_widget_show(beat_window) ;	

	
	
	
	
	return 0;	
}
